<?php

namespace geo\components;

use Faker\Provider\Uuid;
use Yii;
use yii\base\Component;
use yii\helpers\Json;
use yii\httpclient\Client;
use yii\httpclient\CurlTransport;

/**
 * Класс для работы с geo сервисом
 *
 * Class LocalizationComponent
 *
 * @package localization\components
 */
class GeoComponent extends Component
{
    /** @var string $url URL сервиса локализации */
    public $url = '';
    /** @var string $login Логин пользователя */
    public $login = '';
    /** @var string $password Пароль пользователя */
    public $password = '';

    /**
     * @var Client $client
     */
    public $client;

    /** @var string Тип запроса POST */
    const TYPE_REQUEST_POST = 'post';
    /** @var string Тип запроса GET */
    const TYPE_REQUEST_GET = 'get';

    const URL_REQUEST_AUTOCOMPLETE = 'v1/google/autocomplete/';
    const URL_REQUEST_GET_PLACE_DETAIL = 'v1/google/get-place-details/';
    const URL_REQUEST_GET_PLACES_BY_ADDRESS = 'v1/google/get-places-by-address/';
    const URL_REQUEST_GET_PLACES_BY_CORDS = 'v1/google/get-places-by-cords/';
    const URL_REQUEST_GET_PLACES_BY_CORDS_NOT_ONLY_STREET_ADDRESS = 'v1/google/get-places-by-cords-not-only-street-address/';
    const URL_REQUEST_GET_DIRECTION = 'v1/google/get-direction/';

    public function init()
    {
        parent::init();
        $this->client = new Client(
            [
                'transport' => CurlTransport::class,
            ]
        );
    }

    /**
     * Авторизоваться и вернуть токен
     *
     * @return string
     */
    public function auth(): string
    {
        $data = [
            'login' => $this->login,
            'password' => $this->password,
        ];

        $key = $this->url . 'v1/users/auth/auth' . json_encode($data);
        $token = Yii::$app->cache->get($key);
        if (empty($token)) {
            $result = $this->sendRequest(
                $this->url . '/v1/users/auth/auth',
                $data,
                self::TYPE_REQUEST_POST,
                true
            );
            $token = $result['token'];
            Yii::$app->cache->set($key, $token, 3600 * 24);
        }

        return $token;
    }

    /**
     * Произвести запрос в Geo сервис
     *
     * @param string $typeRequest Тип запроса
     * @param array $data данные запроса
     *
     * @return array|mixed Вернёт результат заапрос в GEO сервис
     */
    public function request(string $typeRequest, array $data)
    {
        $data['lang'] = \Yii::$app->language;
        $url = $this->url . $typeRequest;

        return $this->sendRequest($url, $data, self::TYPE_REQUEST_GET);
    }

    /**
     * Отправить запрос в сервис
     *
     * @param string $url URL запроса
     * @param array $data Данные запроса
     * @param string $typeRequest Тип запроса
     * @param bool $isAuth Является ли запрос авторизацией
     *
     * @return array|mixed
     */
    protected function sendRequest($url, $data, $typeRequest = self::TYPE_REQUEST_POST, $isAuth = false)
    {
        $request = $this->client
            ->createRequest()
            ->setMethod($typeRequest)
            ->setUrl($url);

        $xUuid = empty($_ENV['X-UUID']) ? Uuid::uuid() : $_ENV['X-UUID'];
        $headers['X-UUID'] = $xUuid;

        //В зависимости от типа надо добавить поля
        if ($typeRequest === self::TYPE_REQUEST_POST) {
            $headers['content-type'] = 'application/json';
            $request
                ->setFormat(Client::FORMAT_JSON)
                ->setContent(Json::encode($data));
        } else {
            $request->setData($data);
        }

        //Если запрос не на авторизацию, то надо вставить токен
        if (!$isAuth) {
            $headers['Authorization'] = 'Bearer ' . $this->auth();
        }

        $request->addHeaders($headers);

        //Отпавляем запрос и получем ответ
        $response = $request->send();
        $result = $response->getData();

        if (isset($result['error']['code']) && $result['error']['code'] == 4404) {
            return [];
        }

        return $result['result'];
    }

}
