<?php

namespace yandexObjectStorage\components;

use Aws\Credentials\Credentials;
use Aws\Result;
use Aws\S3\Exception\S3Exception;
use Aws\S3\S3Client;
use Exception;

/**
 * Компонент для работы с Yandex Object Storage
 *
 * @package common\components\senders
 */
class YandexObjectStorage
{
    /** @var string */
    public $region;

    /** @var string */
    public $serviceUser;

    /** @var string */
    public $awsAccessKeyId;

    /** @var string */
    public $awsSecretAccessKey;

    /** @var string */
    public $bucket;

    /** @var string */
    public $url;

    /** @var string */
    public $directoryProject;

    /** @var string */
    public $directoryFiles;

    /** @var S3Client $client */
    private $client;

    public function __construct($params)
    {
        $this->url = $params['url'] ?? null;
        $this->bucket = $params['bucket'] ?? null;
        $this->awsSecretAccessKey = $params['awsSecretAccessKey'] ?? null;
        $this->awsAccessKeyId = $params['awsAccessKeyId'] ?? null;
        $this->serviceUser = $params['serviceUser'] ?? null;
        $this->region = $params['region'] ?? null;
        $this->directoryProject = $params['directoryProject'] ?? null;
    }

    public function init(): void
    {
        $this->client = new S3Client(
            [
                'endpoint' => $this->url,
                'version' => 'latest',
                'region' => $this->region,
                'credentials' => new Credentials($this->awsAccessKeyId, $this->awsSecretAccessKey),
            ]
        );
    }

    /**
     * Получить файл из YOS
     *
     * @param string $objectKey Идентификатор объекта в YOS
     * @param string $directory Путь к папке в которую сохраниться файл
     *
     * @return Result
     * @throws Exception
     */
    public function getObject(string $objectKey, $directory): Result
    {
        try {
            $filePath = $this->directoryProject . DIRECTORY_SEPARATOR
                . $this->directoryFiles . DIRECTORY_SEPARATOR . $objectKey;

            $result = $this->client->getObject(
                [
                    'Bucket' => $this->bucket,
                    'Key' => $filePath,
                    'SaveAs' => $directory . $objectKey,
                ]
            );

            return $result;
        } catch (S3Exception $e) {
            throw new Exception($e->getAwsErrorMessage());
        }
    }

    /**
     * Загрузить файл в YOS
     *
     * @param string $fullPathToFile Путь к файлу который нужно передать
     * @param string $objectKey Идентификатор, под которым будет сохранен объект в YOS
     *
     * @return mixed
     * @throws Exception
     */
    public function uploadObject(string $fullPathToFile, string $objectKey): Result
    {
        $filePath = $this->directoryProject . DIRECTORY_SEPARATOR
            . $this->directoryFiles . DIRECTORY_SEPARATOR . $objectKey;

        return $this->client->upload(
            $this->bucket,
            $filePath,
            fopen($fullPathToFile, 'r')
        );
    }
}
