<?php

namespace iwaydev\yii2\log;

use iwaydev\components\helpers\Base;
use Yii;
use yii\base\UnknownPropertyException;
use yii\helpers\VarDumper;
use yii\log\FileTarget;
use yii\helpers\FileHelper;
use yii\log\Logger;
use yii\web\Request;

/**
 * Предоставляет возможность собирать имя файла лога из составляющих, с участием текущей даты.
 * Имя файла генерируется при каждой записи в лог, поэтому если дата изменяется, создается новый актуальный лог
 *
 * @package common\components\log
 */
class ExtendedFileTarget extends FileTarget
{
    const BASE_CATEGORY = 'application';

    public $filePath = null;
    public $filePrefix = null;
    public $fileSuffix = '.log';
    public $traceLevel = 0;
    /** @var bool Если true то в логах время будет указываться с точностью до миллисекунд */
    public $microtime = false;
    public $dirMode = 0777;
    public $fileMode = 0777;

    /**
     * @inheritdoc
     */
    public function init()
    {
        $this->initLogFile();
    }

    /**
     * @inheritdoc
     *
     * Отличается от родителського тем, что исключение при присвоении несуществующего свойства игнорируется,
     * это позволяет нам передавать в конструктор сервиса весь массив входящих данных
     */
    public function __set($name, $value)
    {
        try {
            parent::__set($name, $value);
        } catch (UnknownPropertyException $e) {
        }
    }

    protected function initLogFile()
    {
        if ($this->filePath === null) {
            $this->logFile = Yii::$app->getRuntimePath() . '/logs/app.log';
            $logPath = dirname($this->logFile);
        } else {
            $currentDate = (new \DateTime())->format('Y-m-d');

            // Если забыли дописать дефис в конфигах
            $prefix = ($this->filePrefix === null || substr($this->filePrefix, -1) === '-')
                ? $this->filePrefix
                : ($this->filePrefix . '-');

            $this->logFile = Yii::getAlias($this->filePath)
                . DIRECTORY_SEPARATOR
                . $prefix
                . $currentDate
                . $this->fileSuffix;

            $logPath = Yii::getAlias($this->filePath);
        }

        if (!is_dir($logPath)) {
            FileHelper::createDirectory($logPath, $this->dirMode, true);
        }

        if ($this->maxLogFiles < 1) {
            $this->maxLogFiles = 1;
        }
        if ($this->maxFileSize < 1) {
            $this->maxFileSize = 1;
        }
    }

    public function export()
    {
        $this->initLogFile();
        parent::export();
    }

    protected function rotateFiles()
    {
        $this->initLogFile();
        parent::rotateFiles();
    }

    public function formatMessage($message)
    {
        $usec = '';
        [$text, $level, $category, $timestamp] = $message;
        if ($this->microtime) {
            [$timestamp, $usec] = explode(".", (string)$timestamp);
            $usec = '.' . $usec;
        }
        $level = Logger::getLevelName($level);
        if (!is_string($text)) {
            if ($text instanceof \Exception) {
                $text = (string)$text;
            } else {
                $text = VarDumper::export($text);
            }
        }
        $traces = [];
        if (isset($message[4]) && $this->traceLevel > 0) {
            $count = 0;
            foreach ($message[4] as $trace) {
                $traces[] = "in {$trace['file']}:{$trace['line']}";
                if (++$count >= $this->traceLevel) {
                    break;
                }
            }
        }

        $prefix = $this->getMessagePrefix($message);

        return date('Y-m-d H:i:s', $timestamp) . $usec . " {$prefix}[$level][$category] $text"
            . (empty($traces) ? '' : "\n    " . implode("\n    ", $traces));
    }

    /**
     * Добавляем к префиксу x-uuid
     *
     * @param array $message
     *
     * @return mixed|string
     * @throws \Throwable
     * @throws \yii\base\InvalidConfigException
     */
    public function getMessagePrefix($message)
    {
        $prefix = parent::getMessagePrefix($message);
        $xUuid = Base::getXUuid();

        return $prefix . '[' . $xUuid . ']';
    }
}
