<?php

namespace iwaydev\yii2\log;

use codemix\streamlog\Target;
use common\helpers\ErrorCodeHelper;
use DomainException;
use Exception;
use iwaydev\components\helpers\AbstractErrorCode;
use iwaydev\components\helpers\Base;
use Throwable;
use Yii;
use yii\base\UnknownPropertyException;
use yii\helpers\Json;
use yii\helpers\VarDumper;
use yii\log\Logger;

/**
 * Class StreamTarget
 *
 * @package iwaydev\yii2\log
 */
class StreamTarget extends Target
{
    public $url = 'php://stdout';
    public string $serverName;
    public string $projectName;

    public function init()
    {
        parent::init();

        if (!class_exists('\common\helpers\ErrorCodeHelper')) {
            throw new DomainException('Class \common\helpers\ErrorCodeHelper not found');
        }

        if (!((new ErrorCodeHelper()) instanceof AbstractErrorCode)) {
            throw new DomainException(
                'Class ' . ErrorCodeHelper::class . ' the instance must be a descendant ' .
                AbstractErrorCode::class
            );
        }

        if (empty($this->serverName)) {
            throw new DomainException('"serverName" property cannot be blank');
        }

        if (empty($this->projectName)) {
            throw new DomainException('"projectName" property cannot be blank');
        }
    }

    /**
     * @inheritdoc
     *
     * Отличается от родителського тем, что исключение при присвоении несуществующего свойства игнорируется,
     * это позволяет нам передавать в конструктор сервиса весь массив входящих данных
     */
    public function __set($name, $value)
    {
        try {
            parent::__set($name, $value);
        } catch (UnknownPropertyException $e) {
        }
    }

    public function formatMessage($message)
    {
        $queryParams = [];
        $bodyParams = [];
        [$text, $level, $category, $timestamp] = $message;
        if (!is_string($text)) {
            // exceptions may not be serializable if in the call stack somewhere is a Closure
            if ($text instanceof Throwable || $text instanceof Exception) {
                /** @var Throwable $text */
                $errorCode = ErrorCodeHelper::getCodeProject($text->getCode());
                $trace = $text->getTraceAsString();
                $file = $text->getFile() . '(' . $text->getLine() . ')';
                $text = $text->getMessage();
                $queryParams = Yii::$app->request->getQueryParams();
                $bodyParams = Yii::$app->request->getBodyParams();
            } elseif (is_array($text)) {
                if (isset($text['paramsTrace']) && $text['paramsTrace'] === true) {
                    $queryParams = Yii::$app->request->getQueryParams();
                    $bodyParams = Yii::$app->request->getBodyParams();
                    unset($text['paramsTrace']);
                }
                if ((count($text) == 1) && is_string($text[0])) {
                    $text = $text[0];
                } else {
                    $text = Json::encode($text, JSON_THROW_ON_ERROR | JSON_UNESCAPED_SLASHES | JSON_UNESCAPED_UNICODE);
                }
            } else {
                $text = VarDumper::export($text);
            }
        }

        $data = [
            'entry_point' => Yii::$app->id,
            'ip' => Base::getRealIp() ?? '-',
            'uuid' => Base::getXUuid(),
            'level' => Logger::getLevelName($level),
            'message' => $text,
            'query_params' => $queryParams,
            'body_params' => $bodyParams,
            'error_code' => $errorCode ?? '-',
            'file' => $file ?? '-',
            'trace' => $trace ?? '-',
        ];

        $message = [
            'server' => $this->serverName,
            'project' => $this->projectName,
            'category' => $category,
            'data' => $data
        ];
        $messageJson = Json::encode($message);

        $text = $this->prefixString . $messageJson;

        return $this->replaceNewline === null ?
            $text :
            str_replace("\n", $this->replaceNewline, $text);
    }
}
